import numpy as np
import pandas as pd
from bed_reader import open_bed
import pdb
from matplotlib import pyplot as plt
pd.options.mode.chained_assignment = None

def unit_tester(current_output, known_output_file_name, header_type):

    """
    
    Purpose
    -------
    Given input with known output, theis function compare's each function's current output against the correct output. 

    Parameters
    ----------
    current_output: a functions actual output during unit testing.
    known_output_file_name: name of the file containing the output that the function being tested is expected to have.
    header_type: either 0 (if there is a header) or None (if there is no header). Passed directly to Pandas' read_csv function.

    Returns
    -------
    It returns nothing. It prints whether or not the function passed it's unit test. 
    
    """
    if  known_output_file_name[-4:] == ".bed":
        bed_reader = open_bed("unit_testing_files/correct_write_bed_file_output.bed", count_A1 = True, num_threads = 1)
        correct_output = bed_reader.read(dtype = 'int8')
    elif known_output_file_name[-4:] == ".bim" or known_output_file_name[-4:] == ".fam":
        correct_output = pd.read_csv("unit_testing_files/" + known_output_file_name, delim_whitespace = True, header = None,  dtype = str).to_numpy().astype("str")
    else:
        correct_output = pd.read_csv("unit_testing_files/" + known_output_file_name, delimiter = "\t", header = header_type).to_numpy()
    function_instance = 0
    if len(correct_output.shape) == 2:
        if (correct_output.shape)[1] == 1:
            correct_output = correct_output.reshape(-1)
    if "1" in known_output_file_name or "2" in known_output_file_name:
        function_name = known_output_file_name[8:-12]
        function_instance = known_output_file_name[-5]
    else:
        function_name = known_output_file_name[8:-11]
    if isinstance(current_output, pd.DataFrame):
        current_output = current_output.to_numpy()
    if isinstance(current_output, list):
        current_output = np.array(current_output)
    # Importing the values sometimes creates a small ammount of machine error.
    if known_output_file_name[-4:] == ".bim" or known_output_file_name[-4:] == ".fam":
        if np.any(correct_output != current_output):
            print("The " + function_name + " function's " + known_output_file_name[-4:] + " file output failed its unit test.\n")
        else:
            print("The " + function_name + " function's " + known_output_file_name[-4:] + " file output passed its unit test.\n")
    else:
        if np.any(np.round(correct_output, 12) != np.round(current_output, 12)):
            if function_instance == "1" or function_instance == "2":
                print("Instance " + function_instance + " of the " + function_name + " function failed its unit test.\n")
            else:  
                print("The " + function_name + " function failed its unit test.\n")
        else:
            if function_instance == "1" or function_instance == "2":
                print("Instance " + function_instance + " of the " + function_name + " function passed its unit test.\n")
            else:  
                print("The " + function_name + " function passed its unit test.\n")


def test_drawn_breakpoints(breakpoints, probabilities, chromosome_number, output_plink_filename_prefix):

    """
    
    Purpose
    -------
    To show that each breakpoint is, on average, drawn the expected number of times given the "probabilities" 1D array. 
    To confirms that each row in the "breakpoints" 2D array contains 1 or 0 instances of each breakpoint. 

    Parameters
    ----------
    breakpoints: an NxB numpy array, where N is the number of simulated individuals, and B is the number of breakpoints per chromosome in each individual. 
                 each breakpoint is an index in probabilities, which directly corresponds to a specific recombination interval.
    probabilities: a 1D numpy array. The ith element is the probability of drawing a breakpoint at the ith recombination 
                   interval in a single draw, noting that each row in "breakpoints" cannot contain any breakpoint more than once. 
    chromosome_number: the cromosome being analyzed

    Returns
    -------
    It returns nothing. It plots the expected number of times each breakpoint index is drawn against the actual number of times. 
    for each chromosome. It also prints the row numbers of any breakpoint vectors that contain more than one instance of the same breakpoint. 
    
    """
    
    num_breakpoints = len(breakpoints[0])
    num_samples = len(breakpoints)
    breakpoint_indices = list(range(len(probabilities)))
    breakpoint_counts = dict(zip(breakpoint_indices, len(probabilities)*[0]))

    breakpoint_vecs_with_unwanted_replacement = []
    for i, breakpoint_vec in enumerate(breakpoints):
        if len(breakpoint_vec) != len(np.unique(breakpoint_vec)):
            breakpoint_vecs_with_unwanted_replacement.append(i)
        for b in breakpoint_vec:
            breakpoint_counts[b] += 1

    print("The samples in the following list have drawn the same breakpoint more than once:")
    print(breakpoint_vecs_with_unwanted_replacement)
    if len(breakpoint_vecs_with_unwanted_replacement) == 0:
        print("Every sample has successfully drawn breakpoints without replacement")
    if len(breakpoint_vecs_with_unwanted_replacement) > 0:
        print("Some samples have drawn breakpoints with replacement. This is an error, and it should be attended to.")

    expected_num_breakpoint_counts = []
    num_breakpoint_counts = []
    for index in breakpoint_indices:
        p_not_drawing_breakpoint = (1 - probabilities[index])**num_breakpoints
        p_drawing_breakpoint = 1 - p_not_drawing_breakpoint
        expected_num_breakpoint_counts.append(num_samples*p_drawing_breakpoint)
        num_breakpoint_counts.append(breakpoint_counts[index])

    expected_num_breakpoint_counts = np.array(expected_num_breakpoint_counts)
    num_breakpoint_counts = np.array(num_breakpoint_counts)

    plt.plot(expected_num_breakpoint_counts, num_breakpoint_counts, '*', label = "expected vs actual breakpoint index counts")
    plt.plot(expected_num_breakpoint_counts, expected_num_breakpoint_counts, '-', label = "y = x")
    plt.legend(loc = "upper left")
    plt.xlabel("expected number of breakpoints", fontsize = 14)
    plt.ylabel("number of breakpoints", fontsize = 14)
    if "/" in output_plink_filename_prefix:
        prefix = "/".join((output_plink_filename_prefix.split("/"))[:-1]) + "/"
        print(prefix)
    else:
        prefix = ""
    plt.savefig(prefix + "expected_vs_actual_breakpoint_counts_full_range_chr" + str(chromosome_number) + ".png")
    plt.clf()

def test_breakpoint_SNP_mapping(old_breakpoints, rcmb_interval_positions, new_breakpoints, SNP_positions):

    """
    
    Purpose
    -------
    To confirm for every drawn breakpoint that, when each breakpoint's recombination interval is replaced with the genomic position of a SNP,
    that the SNP's genomic position is in-between the genomic positions of the recombination interval's left and right boundaries.     

    Parameters
    ----------
    old_breakpoints: an NxB numpy array, where N is the number of simulated individuals, and B is the number of breakpoints per chromosome in each individual. 
                     each breakpoint is an index in probabilities, which directly corresponds to a specific recombination interval.
    rcmb_interval_positions: a list where ith index contains the genomic position of the left bound of the recombination interbal, which corresponds 
                             to the ith element in "probabilities." Note that the left bound of the ith interval is the right bound of the (i-1)th interval. 
    new_breakpoints: an NxB numpy array, where N is the number of simulated individuals, and B is the number of breakpoints per chromosome in each individual. 
                     each breakpoint is the row number of a SNP in the input bim file, which directly corresponds to a specific genomic location.
    SNP_positions: a list where ith index contains the genomic position of the SNP in the input bim file's ith row. 

    Returns
    -------
    It returns nothing. It prints a list of the (row, column) coordinates of any new breakpoint with a genomic position that
    does not fall in between the genomic positions of the original recombination interval's left and right boundaries.
    
    """

    num_breakpoints = len(old_breakpoints[0])
    erronious_mappings = []
    for i in range(len(old_breakpoints)):
        for k in range(num_breakpoints):
            rcmb_interval_position = rcmb_interval_positions[old_breakpoints[i][k]]
            SNP_position = SNP_positions[new_breakpoints[i][k]]
            next_rcmb_interval = rcmb_interval_positions[old_breakpoints[i][k] + 1]
            if rcmb_interval_position <= SNP_position and SNP_position <= next_rcmb_interval:
                pass
            else:
                erronious_mappings.append((i,k))
    print("The following rcmb interval indices have been incorrectly mapped to SNP indices")
    print(erronious_mappings)
    if len(erronious_mappings) == 0:
        print("Every index has been mapped correctly")
    if len(erronious_mappings) > 0:
        print("Some SNP indices are not contained in the corresponding rcmb interval. This is an error, and it should be attended to.")
                
def test_simulated_individuals(breakpoints, sampled_individuals, simulated_individuals):
    
    """
    
    Purpose
    -------
    To verify that every simulated individual is comprised of the correct segments from the correct real individuals     

    Parameters
    ----------
    breakpoints: an Nx(B+1) numpy array, where N is the number of simulated individuals, and B is the number of breakpoints per chromosome in each individual. 
                 each breakpoint is the row number of a SNP in the input bim file, which directly corresponds to a specific genomic location. Note that the
                 last column in breakpoints has all elements equal to the current chromosomes last SNP index. The other columns are simple breakpoint values. 
    sampled_individuals: an Nx(B+1)xS numpy array, where N is the number of simulated individuals, B is the number of breakpoints per chromosome 
                         (B breakpoints seperate B+1 segments from B+1 real individuals), and S is the number of SNPs per individual in the current chromosome. 
    simulated_individuals: an NxS numpy arrray. Each array of S SNPs contains concatenated segments from B+1 individuals.  

    Returns
    -------
    It returns nothing. It prints a list of the (row, segment index) of any segments within 
    simulated individuals that do not map back to the correct region of the correct real individual    

    """

    incorrect_segment_coordinates = []
    for j, breakpoint_vec in enumerate(breakpoints):
        startpoint = 0
        for k, breakpoint in enumerate(np.sort(breakpoint_vec)):
            if np.any(sampled_individuals[j][k][startpoint:breakpoint] != simulated_individuals[j][startpoint:breakpoint]):
                incorrect_segment_coordinates.append([j, k])
            startpoint = breakpoint
 
    print("\nThe following segments (simulated_individual row number, segment index) are not from the correct regions of the correct real individuals")
    print(incorrect_segment_coordinates)
    if len(incorrect_segment_coordinates) == 0:
        print("Every segment in every simulated individual is from the correct region of the correct real individual")
    else:
        print("\nThis is a technical error that needs to be addressed\n")


             

        
        
        


import numpy as np
import pandas as pd
import argparse
import os
import pdb
from zipfile import ZipFile
from urllib import request
from bed_reader import open_bed
from bed_reader import to_bed
from matplotlib import pyplot as plt
pd.options.mode.chained_assignment = None

from regens_library import SNP_positions_to_rcmb_intervals
from regens_library import reduce_recomb_rate_info
from regens_library import choice_with_periodic_replacement
from regens_library import centimorgans_to_probabilities
from regens_library import draw_breakpoints
from regens_library import get_samples_fast
from regens_library import breakpoints_to_simulated_individuals
from regens_library import write_bed_file
from regens_library import simulate_phenotypes

def main():

    if not os.path.isdir("input_files"):
        print("\nDownloading input_files for examples and unit testing to your working directory. This may take a moment.\n")
        print("Note: keep the input_files folder in your working directory if you intend to use them\n")
        dataset_url = "https://ndownloader.figshare.com/files/25515740"
        downloaded_path = os.path.join(os.getcwd(), "input_files.zip")
        void =request.urlretrieve(dataset_url, downloaded_path) 
        ZipFile(downloaded_path).extractall()
        os.remove(downloaded_path)
        print("The input_files folder has been downloaded to your working directory\n")

    some_files_present = os.path.isdir("unit_testing_files") and os.path.isdir("runtime_testing")
    other_files_present = os.path.isdir("examples") and os.path.isdir("correctness_testing_GBR") and os.path.isdir("correctness_testing_ACB")
    all_files_present = some_files_present and other_files_present
    if not all_files_present:
        print("Downloading examples and unit testing files to your working directory.\n")
        dataset_url = "https://ndownloader.figshare.com/files/25516322"
        downloaded_path = os.path.join(os.getcwd(), "examples_and_tests_for_REGENS.zip")
        void = request.urlretrieve(dataset_url, downloaded_path) 
        ZipFile(downloaded_path).extractall()
        os.remove(downloaded_path)
        print("The examples and unit testing files have been downloaded to your working directory\n")

    parser = argparse.ArgumentParser()
    parser.add_argument('--in', nargs = "?", type = str, action = "store", dest = "input_plink_filename_prefix", default = "empty", help = 'input plink filename prefix')
    parser.add_argument('--out', nargs = "?", type = str, action = "store", dest = "output_plink_filename_prefix", default = "simulated_whole_genomes", help = 'output plink filename prefix')
    parser.add_argument('--simulate_nbreakpoints', nargs = "?", type = int, action = "store", dest = "num_breakpoints_simulated", default = 4, help = 'If there are n breakpoints, then each simulated chromosome is comprised of n+1 chunks, where each chunk is from a randomly drawn input sample')
    parser.add_argument('--simulate_nsamples', nargs = "?", type = int, action = "store", dest = "num_samples_simulated", default = 1000, help = 'number of simulated')
    parser.add_argument('--phenotype', nargs = "?", type = str, action = "store", dest = "phenotype", default = -9)
    parser.add_argument('--mean_phenotype', nargs = "?", type = float, action = "store", dest = "mean_phenotype", default = None)
    parser.add_argument('--population_code', nargs = "?", type = str, action = "store", dest = "population_code", default = "empty")
    parser.add_argument('--human_genome_version', nargs = "?", type = str, action = "store", dest = "human_genome_version", default = "empty")
    parser.add_argument('--recombination_file_path_prefix', nargs = "?", type = str, action = "store", dest = "recombination_file_path_prefix", default = "empty")
    parser.add_argument('--noise', nargs = "?", type = float, action = "store", dest = "noise", default = 0)
    parser.add_argument('--population_ID', nargs = "?", type = str, action = "store", dest = "population_ID", default = '')
    parser.add_argument('--causal_SNP_IDs_path', nargs = "?", type = str, action = "store", dest = "causal_SNP_IDs_path", default = "empty", help = 'input plink filename prefix')
    parser.add_argument('--major_minor_assignments_path', nargs = "?", type = str, action = "store", dest = "major_minor_assignments_path", default = "standard", help = 'input plink filename prefix')
    parser.add_argument('--betas_path', nargs = "?", type = str, action = "store", dest = "betas_path", default = "empty", help = 'input plink filename prefix')
    parser.add_argument('--SNP_phenotype_map_path', nargs = "?", type = str, action = "store", dest = "SNP_phenotype_map_path", default = "standard", help = 'input plink filename prefix')
    parser.add_argument('--test_functionality', nargs = "?", type = str, action = "store", dest = "test_functionality", default = "no", help = 'input plink filename prefix')

    args = parser.parse_args()
    test_functionality = args.test_functionality
    #imports functions to automatically test various aspects of the code if specified:
    if test_functionality == "test_correctness":
        from regens_testers import test_drawn_breakpoints
        from regens_testers import test_breakpoint_SNP_mapping
        from regens_testers import test_simulated_individuals
    if test_functionality == "test_units":
        from regens_testers import unit_tester
        np.random.seed(0)

    #checks the input plink filename prefix and uses the "ACB" prefix as a default if unit testing
    if test_functionality == "test_units":
        if not (os.path.isfile("input_files/ACB.bed") and os.path.isfile("input_files/ACB.bim") and os.path.isfile("input_files/ACB.fam")):
            print("\nThis program must use input_files/ACB.bed, input_files/ACB.bim and input_files/ACB.fam as input files for unit testing\n")
            exit()
        input_plink_filename_prefix = "input_files/ACB"
        print("\nUsing input_files/ACB.bed, input_files/ACB.bim and input_files/ACB.fam as input files for unit testing\n")
    else:
        input_plink_filename_prefix = args.input_plink_filename_prefix
    bim_file_path = input_plink_filename_prefix + ".bim"
    if os.path.isfile(input_plink_filename_prefix + ".bim") and os.path.isfile(input_plink_filename_prefix + ".bed") and os.path.isfile(input_plink_filename_prefix + ".fam"):
        bim_file = pd.read_csv(bim_file_path, delim_whitespace = True, header = None)
        if len(bim_file.columns) != 6:
            print("\nerror: bim file must be in standard format (it must have 6 columns).\n")
            exit()
        chromosomes = np.unique(bim_file[0])
        bim_SNP_pos_chr = [((bim_file[bim_file[0] == chr])[3]).to_numpy() for chr in chromosomes]
        bim_SNP_complete_pos_chr = [((bim_file[bim_file[0] == chr])[[0, 2, 3]]).to_numpy() for chr in chromosomes]
        bim_SNP_nucleotides_chr = [((bim_file[bim_file[0] == chr])[[4, 5]]).to_numpy().astype('U') for chr in chromosomes]
        bim_SNP_names = (bim_file[1]).to_numpy().astype('U')
        bim_SNP_names_chr = [((bim_file[bim_file[0] == chr])[1]).to_numpy().astype('U') for chr in chromosomes]
        chr_SNP_counts = [len(bim_SNP_pos_chr[i]) for i in range(len(chromosomes))]
        cumulative_SNP_counts = np.cumsum([0] + chr_SNP_counts)
    else:
        print("\nerror: one or more of the three necessary plink files with the filename prefix '" + input_plink_filename_prefix + "' is not present.\n")
        exit()

    # The output filename prefix does not need to be checked. 
    output_plink_filename_prefix = args.output_plink_filename_prefix

    # Checks the number of breakpoints
    num_breakpoints = args.num_breakpoints_simulated
    if num_breakpoints < 0:
        print("\nerror: the number of breakpoints cannot be negative.\n")
        exit()
    if num_breakpoints == 0:
        print("\nwarning: setting 0 breakpoints means that each whole chromosome will be a real individual's whole chromosome from the dataset.\n")
    if num_breakpoints > 5:
        print("\nwarning: you have selected " + str(num_breakpoints) + " breakpoints per chromosome, and there is no good reason to choose more than a few breakpoints per chromosome.\n") 
        print("The simulation may become worse if there aren't roughly as many high recombination rate regions as there are breakpoints, and it may require large ammounts of time and memory.\n")

    # checks the specified phenotype category
    phenotype = args.phenotype
    if phenotype not in ["continuous", "binary", -9]:
        print("\nerror: if a phenotype is specified, it must be 'continuous' or 'binary'.\n")
        exit()

    # checks that the mean phenotype value makes sense
    mean_phenotype = args.mean_phenotype
    if phenotype == "binary":
        if mean_phenotype <= 0 or mean_phenotype >= 1:
            print("The mean phenotype value for a binary phenotype must be greater than 0 ort less than 1")
            exit()

    # checks path information related to the recombination rate info file
    if test_functionality == "test_units":
        print("\nUsing population code ACB for unit testing\n")
        pop_code = "ACB"
    else:
        pop_code = args.population_code
    hg_version = args.human_genome_version
    if (not (os.path.isdir("hg19") and os.path.isdir("hg38"))) and (hg_version == "hg19" or hg_version == "hg38"):
        print("Downloading recombination map files contained in hg19 and hg38 to your working directory. This may take a moment.\n")
        print("Note: When using regens' --human_genome_version flag, keep the hg19 and hg38 folders in your working directory.\n")
        dataset_url = "https://ndownloader.figshare.com/articles/13210796/versions/1"
        downloaded_path = os.path.join(os.getcwd(), "recombination_maps.zip")
        zipped_hg19_path = os.path.join(os.getcwd(), "hg19.zip")
        zipped_hg38_path = os.path.join(os.getcwd(), "hg38.zip")
        void =request.urlretrieve(dataset_url, downloaded_path) 
        ZipFile(downloaded_path).extractall()
        ZipFile(zipped_hg19_path).extractall()
        ZipFile(zipped_hg38_path).extractall()
        os.remove(downloaded_path)
        os.remove(zipped_hg19_path)
        os.remove(zipped_hg38_path)
        print("Folders hg19 and hg38 have been downloaded to your working directory\n")
    rcmb_path_prefix = args.recombination_file_path_prefix
    hg_pop_codes = ["ACB", "ASW", "BEB", "CDX", "CEU", "CHB", "CHS", "CLM", "ESN", "FIN", "GBR", "GIH", "GWD", "IBS", "ITU", "JPT", "KHV", "LWK", "MSL", "MXL", "PEL", "PJL", "PUR", "STU", "TSI", "YRI"]
    
    if rcmb_path_prefix != "empty":
        rcmb_rate_info_path_prefix = rcmb_path_prefix
    elif hg_version in ["hg19", "hg38"]:
        if pop_code in hg_pop_codes:
            rcmb_rate_info_path_prefix = hg_version + "/" + pop_code + "/" + pop_code 
            rcmb_rate_info_path_prefix += "_recombination_map_hapmap_format_" + hg_version + "_chr_"
            try:
                rcmb_rate_info_file_names = np.sort(os.listdir(hg_version + "/" + pop_code))
                correct_info_file_names = np.sort([pop_code + "_recombination_map_hapmap_format_" + hg_version + "_chr_" + str(i) + ".txt.gz" for i in range(1, 23)])
            except:
                print("\nerror, possibility 1: Your working directory needs to be in the same directory as the triadsim_main.py file.\n")
                print("error, possibility 2: the triadsim_main.py file needs to be in the same directory as the hg19 and hg38 folders.\n") 
                exit()
            try: 
                if np.any(rcmb_rate_info_file_names != correct_info_file_names):
                    exit()
            except:
                print("\nerror: a recombination rate map file has been deleted or is incorrectly named in" + rcmb_rate_info_path_prefix + ".\n")
                exit()        
    else:
        print("\nerror: The human genome version must be 19 or 38, and the population code must be one of the following\n")
        print(hg_pop_codes)
        print("\nAlternatively, you can override the human genome version and population code arguments by adding a path to a custom recombination file with the recombination_file_path_prefix argument\n")
        exit()

    # checks the number of samples to be simulated
    simulation_sample_size = args.num_samples_simulated
    if simulation_sample_size <= 0:
        print("\nerror: the number of simulated samples cannot be negative or 0.\n")
        exit()
    SNP_count = cumulative_SNP_counts[-1]
    minutes = str((126/60)*(simulation_sample_size/10000)*(SNP_count/531170))
    max_ram = str((4.1)*(simulation_sample_size/10000)*(SNP_count/531170))
    print("\nSimulating " + str(simulation_sample_size) + " samples and " + str(SNP_count) + " SNPs will require roughly " + minutes + " minutes and a maximum ram usage of " + max_ram + " GB.\n")

    # This downloads dataframes containing genomic intervals and recombination rates (in CentiMorgans/Megabase). There is one dataframe per chromosome. 
    rcmb_rate_info_paths = [rcmb_rate_info_path_prefix + str(chr) + '.txt.gz' for chr in chromosomes]
    rcmb_rate_info_chr = [pd.read_csv(path, delim_whitespace = True, header = 0, compression = "gzip") for path in rcmb_rate_info_paths]
    if test_functionality == "test_units":
        rcmb_rate_info_chr = [reduce_recomb_rate_info(rcmb_rate_info_chr[i], bim_SNP_pos_chr[i], test_functionality) for i in [0]]
    else:
        rcmb_rate_info_chr = [reduce_recomb_rate_info(rcmb_rate_info_chr[i], bim_SNP_pos_chr[i], test_functionality) for i in range(len(chromosomes))]
    rcmb_rate_intervals_chr = [((rcmb_rate_info_chr[i])["Position(bp)"]).to_numpy().reshape(-1) for i in range(len(rcmb_rate_info_chr))]

    # checks that percent noise value makes sense
    noise = args.noise
    if noise < 0:
        print("The ammount of noise as a percentage of the average signal cannot be negative.")
        exit()

    # gets the population ID for file_names. Doesn't need a type check. 
    population_ID = args.population_ID
    if population_ID == '' and pop_code in hg_pop_codes:
        population_ID = pop_code

    # checks to make sure that input phenotype simulation files exist in the working directory
    causal_SNP_IDs_path = args.causal_SNP_IDs_path
    major_minor_assignments_path = args.major_minor_assignments_path
    betas_path = args.betas_path
    SNP_phenotype_map_path = args.SNP_phenotype_map_path
    if not (os.path.isfile(causal_SNP_IDs_path) or (causal_SNP_IDs_path == "empty" and phenotype == -9)):
        print("\nerror: A causal SNP IDs file named '" + causal_SNP_IDs_path + "' does not exist in your working directory.\n")
        exit()
    if not (os.path.isfile(betas_path) or (betas_path == "empty" and phenotype == -9)):
        print("\nerror: A betas file named '" + betas_path + "' does not exist in your working directory.\n")
        exit()
    if not (os.path.isfile(major_minor_assignments_path) or (major_minor_assignments_path == "standard")):
        print("\nerror: A major minor assignments file named '" + major_minor_assignments_path + "' does not exist in your working directory.\n")
        exit()
    if not (os.path.isfile(SNP_phenotype_map_path) or (SNP_phenotype_map_path == "standard")):
        print("\nerror: A SNP phenotype map file named '" + SNP_phenotype_map_path + "' does not exist in your working directory.\n")
        exit()

    # runs the main algorithm. 
    if test_functionality == "test_units":
        chromosomes = chromosomes[0:1]
    for i in range(len(chromosomes)):

        print("simulating chromosome " + str(chromosomes[i]))

        input_sample_column_bounds = [cumulative_SNP_counts[i], cumulative_SNP_counts[i + 1]]
        samples = get_samples_fast(simulation_sample_size,
                                   input_sample_column_bounds, 
                                   input_plink_filename_prefix, 
                                   num_breakpoints,
                                   test_functionality)

        breakpoints = draw_breakpoints(rcmb_rate_info_chr[i], 
                                       bim_SNP_pos_chr[i], 
                                       num_breakpoints, 
                                       simulation_sample_size,
                                       test_functionality,
                                       (i + 1),
                                       output_plink_filename_prefix)

        simulated_individuals = breakpoints_to_simulated_individuals(breakpoints, 
                                                                     chr_SNP_counts[i], 
                                                                     samples,
                                                                     test_functionality)

        write_bed_file(simulated_individuals, 
                       bim_SNP_names_chr[i], 
                       output_plink_filename_prefix + "_chr" + str(chromosomes[i]) + ".bed",
                           bim_SNP_complete_pos_chr[i],
                       bim_SNP_nucleotides_chr[i],
                       population_ID,
                       test_functionality)

    if (phenotype == "binary" or phenotype == "continuous") and test_functionality != "test_units":

        sample_size = len(simulated_individuals)
        output_file_names = [output_plink_filename_prefix + "_chr" + str(chromosomes[i]) + ".bed" for i in range(len(chromosomes))]

        simulated_phenotypes = simulate_phenotypes(output_file_names, causal_SNP_IDs_path, cumulative_SNP_counts,
                                                   major_minor_assignments_path, betas_path, mean_phenotype, sample_size,
                                                   bim_SNP_names, phenotype, SNP_phenotype_map_path, noise = noise)
    
        if phenotype == "binary":
            legend_label = "phenotype barplot (mean = " + str(np.round(np.mean(simulated_phenotypes), 5))
            legend_label += ", min = " + str(np.round(np.min(simulated_phenotypes), 5)) + ", max = " + str(np.round(np.max(simulated_phenotypes), 5)) + ")"
            plt.bar([0, 1], [np.sum(simulated_phenotypes == 0), np.sum(simulated_phenotypes == 1)], label = legend_label, tick_label = ["0", "1"])
        if phenotype == "continuous":
            legend_label = "phenotype value histogram (mean = " + str(np.round(np.mean(simulated_phenotypes), 5))
            legend_label += ", min = " + str(np.round(np.min(simulated_phenotypes), 5)) + ", max = " + str(np.round(np.max(simulated_phenotypes), 5)) + ")"
            plt.hist(simulated_phenotypes, bins = 30, label = legend_label)
        plt.legend()
        plt.xlabel("phenotype value", fontsize = 12)
        plt.ylabel("number of value instances", fontsize = 12)
        plt.savefig(output_plink_filename_prefix + "_phenotype_profile.png")
        plt.clf()

        for file_name in output_file_names:
            fam_file_name = file_name[:-3] + "fam"
            fam_file = pd.read_csv(fam_file_name, delimiter = " ", header = None)
            fam_file[5] = simulated_phenotypes.reshape(-1,1)
            fam_file[4] = 2
            fam_file.to_csv(fam_file_name, sep = " ", header = False, index = False)

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import pdb
from time import time
from bed_reader import open_bed
from bed_reader import to_bed
from scipy.optimize import root
from copy import deepcopy as COPY
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import LinearRegression
from regens_testers import unit_tester
pd.options.mode.chained_assignment = None

# TODO: double check that the modified recombination_rate_to_probabilities function works as expected. 
# TODO: double check why you don't subtract 1 from the first "SNP_positions_to_rcmb_intervals" use but you do from the second. 
# IMPORTANT DEFINITION: a breakpoint refers to the genomic position of the boundary between segments. 
# from two sampled whole genomes. It does not refer to the programming definition of breakpoint.  

def reduce_recomb_rate_info(rcmb_rate_info, bim_SNP_positions, test_functionality):

    """
    
    Purpose
    -------
    To reduce the number of rows in rcmb_rate_info so that every resulting genomic interval contains at least one SNP. 

    Parameters
    ----------
    rcmb_rate_info: a pandas dataframe with two columns. "Position(bp)" is the genomic position of the ith interval's left
                    boundary. The ith row is both the left boundry of the ith interval and the right boundary of the (i-1)th 
                    interval. "Map(cM)" is the cumulative recombination rate in centiMorgans, which is 0 in the first row.

    bim_SNP_positions: the genomic positions of every SNP directly from the input bim file. 

    Returns
    -------
    reduced_rcmb_rate_info: "rcmb_rate_info" with rows removed to ensure that at least one 
                            SNP position from "bim_SNP_positions" resides in every interval. 
    
    """
    
    rcmb_rate_intervals = rcmb_rate_info["Position(bp)"].to_numpy()

    # INDEXING NOTE: Not subtracting one from "SNP_pos_rcmb_interval_map" returns the closest indices of "rcmb_rate_intervals"
    #                boundaries at genomic positions to the RIGHT of each SNP's genomic position in bim_SNP_positions.
    #                The only row still needed is the closest boundary to the LEFT of the first SNP. This is aquired by
    #                implementing "occupied_rcmb_intervals[np.min(np.where(occupied_rcmb_intervals == True)) - 1] = True".

    SNP_pos_rcmb_interval_map = SNP_positions_to_rcmb_intervals(rcmb_rate_intervals, COPY(bim_SNP_positions), test_functionality, context = 1)
    all_rcmb_intervals = np.arange(len(rcmb_rate_intervals))
    occupied_rcmb_intervals = np.isin(all_rcmb_intervals, SNP_pos_rcmb_interval_map)
    occupied_rcmb_intervals[np.min(np.where(occupied_rcmb_intervals == True)) - 1] = True
    reduced_rcmb_rate_info = rcmb_rate_info[occupied_rcmb_intervals]

    #reduced_rcmb_rate_info.to_csv("correct_reduce_recomb_rate_info_output.txt", sep = "\t", header = True, index = False)
    if test_functionality == "test_units":
        unit_tester(reduced_rcmb_rate_info, "correct_reduce_recomb_rate_info_output.txt", 0)
    return(reduced_rcmb_rate_info)

def SNP_positions_to_rcmb_intervals(rcmb_interval_boundaries, bim_SNP_positions, test_functionality, context):

    """
    
    Purpose
    -------
    To produce a list with indices that correspond to "bim_SNP_positions" indices 
    and values that correspond to "rcmb_rate_intervals" indices. This is an intermediate 
    step in converting sampled breakpoint intervals to SNPs from the input dataset.

    Parameters
    ----------
    rcmb_interval_boundaries: The "Position(bp)" column from "rcmb_rate_info". The ith element is the genomic position of
                              both the left boundry of the ith interval and the right boundary of the (i-1)th interval.

    bim_SNP_positions: the genomic positions of every SNP directly from the input bim file.

    Returns
    -------
    SNP_rcmb_interval_indices: a list with indices that correspond to "rcmb_rate_intervals" 
    indices and values that correspond to "bim_SNP_positions" indices    

    """

    prev_start_pos = 0
    SNP_rcmb_interval_indices = np.zeros(len(bim_SNP_positions), dtype = np.int64) 
    too_large_indices = bim_SNP_positions >= np.max(rcmb_interval_boundaries)
    too_small_indices = bim_SNP_positions <= np.min(rcmb_interval_boundaries)
    bim_SNP_positions[too_large_indices] = np.max(rcmb_interval_boundaries) - 1
    bim_SNP_positions[too_small_indices] = np.min(rcmb_interval_boundaries) + 1
    num_outer_SNPs = np.sum(too_large_indices) + np.sum(too_small_indices) 
    if num_outer_SNPs > 0:
        print("\nCAUTION: " + str(num_outer_SNPs) + " SNP(s) in the bim file are outside of the positional range considered by the recombination rate info file.\n")
        print("This is unlikely to be an issue if the number of such SNPs is small. Otherwise, remove those SNPs from the input plink files\n")
        print("bim file SNPs earlier than the first recombination interval are assumed to be inside of the first recombination interval.\n")
        print("bim file SNPs farther than the last recombination interval are assumed to be inside of the last recombination interval.\n")
    for i in range(len(bim_SNP_positions)):
        for k in range(len(rcmb_interval_boundaries) - prev_start_pos):
            if rcmb_interval_boundaries[prev_start_pos + k] >= bim_SNP_positions[i]:
                SNP_rcmb_interval_indices[i] = prev_start_pos + k
                prev_start_pos += k
                break
    
    if test_functionality == "test_units":
        unit_tester(SNP_rcmb_interval_indices, "correct_SNP_positions_to_rcmb_intervals_output" + str(context) + ".txt", None)

    return(SNP_rcmb_interval_indices)

def choice_with_periodic_replacement(length, width, probabilities, test_functionality):

    """
    
    Purpose
    -------
    This function's operations are equivalent to the following procedure:
        1) Sample B indices from a probability vector without replacement. 
        2) Replace the sampled indices. 
        3) Repeat Steps 1 and 2 N times. 
    Here, B is the number of breakpoints per chromosome, and N is the number of individuals to be simulated. 

    Parameters
    ----------
    length: The (int) number of individuals to be simulated (N)
    width: the (int) number of breakpoints per chromosome (B)
    probabilities: A numpy array, where the ith element is the probability of sampling the ith 
                   recombination interval from "reduced_rcmb_rate_info" to contain a breakpoint

    Returns
    -------
    an NxB numpy array containing N sets of B recombination interval indices. All indices 
    have 1 subtracted from them because a [0] element is appended onto the beginning of
    the cumulative probability distribution for technical convenience, which increases 
    all of the indices by one higher than rest of the code expects them to be.   

    """

    samples = np.zeros((width, length)).astype(int)
    cumulative_probabilities = np.repeat([np.cumsum(np.append([0], probabilities))], length, axis = 0)
    max_cumulative_vals = cumulative_probabilities[:, -1]
    for i in range(width):
        adjusted_uniform_sample = max_cumulative_vals*np.random.rand(length)
        samples[i] = np.array([np.searchsorted(probs, sample) for probs, sample in zip(cumulative_probabilities, adjusted_uniform_sample)])
        if i < (width - 1):
            for j, pos in enumerate(samples[i]):
                cumulative_probabilities[j][pos:] -= cumulative_probabilities[j][pos] - cumulative_probabilities[j][pos - 1]
            max_cumulative_vals = cumulative_probabilities[:, -1]
    
    if test_functionality == "test_units":
        unit_tester(samples.T - 1, "correct_choice_with_periodic_replacement_output.txt", None)

    return(samples.T - 1)
     
def centimorgans_to_probabilities(recomb_rate_info, test_functionality):

    """
    
    Purpose
    -------
    Convert recombination rates (in cM) to the probability that a recombination event exists at the 
    corresponding interval given that one recombination event exists in the chromosome. This is used 
    as the probability of drawing that interval's genomic position to be the site of a breakpoint. 

    Parameters
    ----------
    recomb_rate_info: Output from the "reduce_recomb_rate_info" function.
    
    Returns
    -------
    a numpy array of probabilities that sum to 1. The ith probability is the probability 
    of drawing the ith recombination interval to use as the location of a breakpoint.

    """ 

    new_cumulative_cM = recomb_rate_info["Map(cM)"].to_numpy()
    rcmb_rates = new_cumulative_cM[1:] - new_cumulative_cM[0:-1]
    expected_Ri = (1 - np.exp(-rcmb_rates/50))/2
    expected_R = np.sum(expected_Ri)
    p_Ri_true_given_R_one = expected_Ri/expected_R

    if test_functionality == "test_units":
        unit_tester(p_Ri_true_given_R_one, "correct_centimorgans_to_probabilities_output.txt", None)

    return(p_Ri_true_given_R_one)

def draw_breakpoints(rcmb_rate_info, bim_SNP_positions, num_breakpoints, simulation_sample_size, test_functionality, chromosome_number, output_plink_filename_prefix):
    
    """

    Purpose
    -------
    Computes breakpoint sampling probabilities with "centimorgans_to_probabilities", Draws breakpoints 
    with "choice_with_periodic_replacement", and converts the breakpoints' corresponding recombination
    interval indices into the indices of input SNPs that reside inside of the recombination interval.    

    Parameters
    ----------
    rcmb_rate_info:  Output from the "reduce_recomb_rate_info" function.
    bim_SNP_positions: the genomic positions of every SNP directly from the input bim file.
    num_breakpoints: user-specified (int) number of breakpoints per chromosome.
    simulation_sample_size: user-specified number of samples to be simulated.
    test_functionality: an argument which, if equal to "yes", tests regens' functionality. It substantially increases runtime. 
    chromosome_number: the chromosome that is currently being simulated. 
    output_plink_filename_prefix: plink prefix of the (bed, bim, fam) fileset that will contain simulated individuals. 
    
    Returns
    -------
    an NxB numpy array containing N sets of B recombination interval indices. 
    Each index is an input SNP's bim row index (also it's bed column index). 

    """ 

    if test_functionality == "test_correctness":
        from regens_testers import test_drawn_breakpoints
        from regens_testers import test_breakpoint_SNP_mapping
      
    SNP_count = len(bim_SNP_positions)
    probabilities = centimorgans_to_probabilities(rcmb_rate_info, test_functionality)
    rcmb_rate_intervals = rcmb_rate_info["Position(bp)"].to_numpy()
    breakpoints = choice_with_periodic_replacement(simulation_sample_size,  
                                                   num_breakpoints, 
                                                   probabilities,
                                                   test_functionality)

    if test_functionality == "test_correctness":
        test_drawn_breakpoints(breakpoints, probabilities, chromosome_number, output_plink_filename_prefix)
        old_breakpoints = COPY(breakpoints)

    # INDEXING NOTE: Subtracting one from "SNP_pos_rcmb_interval_map" returns the closest indices of "rcmb_rate_intervals"
    #                boundaries at genomic positions to the LEFT of each SNP's genomic position in bim_SNP_positions.
    #                This is because all SNPs up to the SNP immediately to the left of the ith breakpoint comprise the ith 
    #                segment, noting that the (B+1)th includes all SNPs after the Bth breakpoint (there are B breakpoints). 

    SNP_pos_rcmb_interval_map = SNP_positions_to_rcmb_intervals(rcmb_rate_intervals, COPY(bim_SNP_positions), test_functionality, context = 2) - 1
   
    rcmb_interval_SNP_pos_map = {}
    for rcmb_interval in np.unique(SNP_pos_rcmb_interval_map):
        rcmb_interval_SNP_pos_map[rcmb_interval] = np.where(SNP_pos_rcmb_interval_map == rcmb_interval)[0]
        
    for jj in range(len(breakpoints)):
        for k in range(num_breakpoints):
            interval_index = breakpoints[jj][k]
            SNP_indices = rcmb_interval_SNP_pos_map[interval_index]
            if len(SNP_indices) == 1:
                breakpoints[jj][k] = SNP_indices[0]
            else:
                breakpoints[jj][k] = SNP_indices[int(len(SNP_indices)*np.random.rand() - 0.5)]

    if test_functionality == "test_correctness":
        test_breakpoint_SNP_mapping(old_breakpoints, rcmb_rate_intervals, breakpoints, bim_SNP_positions)
    if test_functionality == "test_units":
        unit_tester(breakpoints, "correct_draw_breakpoints_output.txt", None)

    return(breakpoints)

def get_samples_fast(simulation_sample_size, bed_col_bounds, plink_file_name_prefix, num_breakpoints, test_functionality):

    """

    Purpose
    -------
    Imports real whole genomes from the input bed file in proportions as close as possible to that of the bed file's population and 
    randomly assigns every breakpoint seperated segment to be copied from one of the imported whole genomes without replacement.    

    Parameters
    ----------
    simulation_sample_size: the (int) number of samples to be simulated (N)
    bed_col_bounds: first and last column indices of all SNP indices that comprise the chromosome being simulated
    num_breakpoints: user-specified (int) number of breakpoints per chromosome (B).
    plink_file_name_prefix: plink prefix of the (bed, bim, fam) fileset that contains real whole genomes. 
    
    Returns
    -------
    an Nx(B+1)xS numpy array containing N sets of B whole chromosomes, each of which have S snps. 
    Each row in the ith (B+1)xS subset will contribute one genomic segment. Those (B+1) genomic 
    segments will be concatenated to comprise the ith simulated individual.  

    """ 

    bed_file_path = plink_file_name_prefix + ".bed"
    bed_reader = open_bed(bed_file_path, count_A1 = True, num_threads = 1)
    num_repeats = int(simulation_sample_size*(num_breakpoints + 1)/int(bed_reader.iid_count) + 1)
    bed_row_indices = np.repeat(range(bed_reader.iid_count), num_repeats)
    np.random.shuffle(bed_row_indices)
    bed_row_indices = bed_row_indices[:simulation_sample_size*(num_breakpoints + 1)]
    bed_file_samples = bed_reader.read((bed_row_indices, slice(bed_col_bounds[0], bed_col_bounds[1])), dtype = 'int8')
    bed_file_samples_dimensions = (int(len(bed_row_indices)/(num_breakpoints + 1)),  
                                   num_breakpoints + 1, 
                                   len(bed_file_samples[0]))
    reshaped_output = bed_file_samples.reshape(bed_file_samples_dimensions)

    if test_functionality == "test_units":
        # gives 3 different counts of how many times each minor allele is drawn. They're small and vanishingly unlikely to all be made correctly by chance. 
        SNP_minor_allele_counts = np.sum(reshaped_output, axis = (0, 1))
        simulated_individual_minor_allele_counts = np.sum(reshaped_output, axis = (1, 2))
        breakpoint_interval_minor_allele_counts = np.sum(reshaped_output, axis = (0, 2))
        unit_tester(SNP_minor_allele_counts, "correct_get_samples_fast_SNP_minor_allele_counts_output.txt", None)
        unit_tester(simulated_individual_minor_allele_counts, "correct_get_samples_fast_simulated_individual_minor_allele_counts_output.txt", None)
        unit_tester(breakpoint_interval_minor_allele_counts, "correct_get_samples_fast_breakpoint_interval_minor_allele_counts_output.txt", None)

    return(reshaped_output)

def breakpoints_to_simulated_individuals(breakpoints, SNP_count, sampled_individuals, test_functionality):

    """

    Purpose
    -------
    Imports real whole genomes from the input bed file in proportions as close as possible to that of the bed file's population and 
    randomly assigns every breakpoint seperated segment to be copied from one of the imported whole genomes without replacement.    

    Parameters
    ----------
    sampled_individuals: an Nx(B+1)xS numpy array containing N sets of B whole chromosomes, each of which have S snps. 
                         Each row in the ith (B+1)xS subset will contribute one genomic segment. Those (B+1) genomic 
                         segments will be concatenated to comprise the ith simulated individual.  
    breakpoints: an Nx(B+1) numpy array, where N is the number of simulated individuals, and B is the number of breakpoints 
                 per chromosome in each individual. Each breakpoint is the row number of a SNP in the input bim file, 
                 which directly corresponds to a specific genomic location.
    SNP_count: The number of input SNPs in the chromosome currently being simulated 
    
    Returns
    -------
    an NxS numpy array containing N simulated chromosomes, each of which has segments from B+1 individuals and S total SNPs.
    """ 

    breakpoints = np.append(breakpoints, SNP_count*np.ones((len(breakpoints), 1), dtype = np.int64), axis = 1)
    breakpoints_masks = np.zeros((len(breakpoints), len(breakpoints[0]), SNP_count), dtype = bool)
    for j, breakpoint_vec in enumerate(breakpoints):
        startpoint = 0
        for k, breakpoint in enumerate(np.sort(breakpoint_vec)):
            breakpoints_masks[j][k][startpoint:breakpoint] = True
            startpoint = breakpoint
    simulated_individuals = sampled_individuals[breakpoints_masks].reshape(-1, SNP_count)

    if test_functionality == "test_correctness":
        from regens_testers import test_simulated_individuals
        test_simulated_individuals(breakpoints, sampled_individuals, simulated_individuals)

    return(simulated_individuals)

def write_bed_file(simulated_individuals, bim_SNP_names, output_name, bim_SNP_complete_pos, bim_SNP_nucleotides, population_ID, test_functionality):

    """

    Purpose
    -------
    to write the simulated data into realistic (bed, bim, fam) filesets. Does not include phenotypes at this point.      

    Parameters
    ----------
    sampled_individuals: an Nx(B+1)xS numpy array containing N sets of B whole chromosomes, each of which have S snps. 
                         Each row in the ith (B+1)xS subset will contribute one genomic segment. Those (B+1) genomic 
                         segments will be concatenated to comprise the ith simulated individual.  
    bim_SNP_names: a list of SNP's rsIDs from the input bed file. 
    output_name: name of the output bed file, which annotates the chromosome that it belongs to. 
    bim_SNP_complete_pos: Sx3 numpy array. Columns comprise the first, third, and fourth columns from the input bim file. 
    bim_SNP_nucleotides: Sx2 numpy array. Columns comprise the fifth and sixth columns from the input bim file (i.e. major and minor alleles).
                         CAUTION: minor alleles with frequencies near 50% may become the major allele after the simulation
                         because the simulated allele frequency always deviates from the real allele frequency by a small ammout.
                         This makes plink flip the sign of r values for simulated SNP pairs relative to real SNP pairs
                         if plink's --keep-allele-order flag is not used when computing the r values with plink. 
    population_ID: An input argument that is concatenated to each sample's row index to comprise columns 1 and 2 for the output fam file.
                   If no input argument is selected, then it includes the popilation ID from the 1000 genomes input plink fileset. If 
                   the input plink files are custom, then it includes an empty string as the population_ID. 

    Returns
    -------
    It returns nothing. It only writes the simulated data into plink files. 

    """ 

    simulated_IDs = np.array([population_ID + "_" + str(i) for i in range(1, len(simulated_individuals) + 1)])
    metadata = {"fid": simulated_IDs,
                "iid": simulated_IDs,
                "sex": np.array([2]*len(simulated_IDs)),
                "pheno": np.array([-9]*len(simulated_IDs)),
                "chromosome": bim_SNP_complete_pos.T[0],
                "sid": bim_SNP_names,
                "cm_position": bim_SNP_complete_pos.T[1],
                "bp_position": bim_SNP_complete_pos.T[2],
                "allele_1": bim_SNP_nucleotides.T[0],
                "allele_2": bim_SNP_nucleotides.T[1]}

    to_bed(output_name, simulated_individuals, properties = metadata, count_A1 = True)

    if test_functionality == "test_units":
        bed_reader = open_bed(output_name, count_A1 = True, num_threads = 1)
        output_bed_file = bed_reader.read(dtype = 'int8')
        output_bim_file = pd.read_csv(output_name[:-4] + ".bim", delimiter = "\t", header = None, dtype = str).to_numpy().astype("str")
        output_fam_file = pd.read_csv(output_name[:-4] + ".fam", delimiter = " ", header = None, dtype = str).to_numpy().astype("str")
        unit_tester(output_bed_file, "correct_write_bed_file_output.bed", None)
        unit_tester(output_bim_file, "correct_write_bed_file_output.bim", None)
        unit_tester(output_fam_file, "correct_write_bed_file_output.fam", None)

        
def get_feature_SNPs(feature_SNP_IDs, cumulative_SNP_counts, output_file_names, sample_size, bim_SNP_names):

    """

    Purpose
    -------
    Each row in the causal_SNP_IDs contains one or more SNPs that transform to compute a single genetic feature. 
    The purpose of this function is to return all of those SNPs' genotypes for one row in the causal_SNP_IDs file.

    Parameters
    ----------
    feature_SNP_IDs: A list of rsIDs from the input plink file (same as the output bim file) that comprise a single feature. 
    cumulative_SNP_counts: the cumulative number of SNPs summed from chromosome 1 to chromosome 22 in ascending order.
    output_file_names: the names of the output bed files for all chromosomes.
    sample_size: the number of samples that have been simulated.
    bim_SNP_names: a list of SNPs from the output bim file (same as the input bim file).

    Returns
    -------
    It returns a 2D numpy array where each column is the column contains the causal genotypes for one SNP. 
    """ 
    
    feature_SNPs = np.zeros((sample_size, len(feature_SNP_IDs)))
    for i, SNP_ID in enumerate(feature_SNP_IDs):
        index = np.where(SNP_ID == bim_SNP_names)[0][0]
        output_file_index = np.min(np.where(index <= cumulative_SNP_counts)) - 1
        output_file_reader = open_bed(output_file_names[output_file_index], count_A1 = True, num_threads = 1)
        adjusted_SNP_index = index - cumulative_SNP_counts[output_file_index]
        feature_SNPs[:, [i]] += output_file_reader.read(index = adjusted_SNP_index, dtype = 'float32')

    return(feature_SNPs)

def simulate_phenotypes(output_file_names, causal_SNP_IDs_path, cumulative_SNP_counts, 
                        major_minor_assignments_path, betas_path, mean_phenotype, sample_size,
                        bim_SNP_names, phenotype, SNP_phenotype_map_path, noise = 0):

    """

    Purpose
    -------
    to simulate correlations between the genotypes of selected SNPs and a continuous or binary phenotype.      

    Parameters
    ----------
    output_file_names: the names of the output bed files for all chromosomes.
    causal_SNP_IDs_path: the path to the file containing causal rsIDs from which phenotype values are simulated. 
    cumulative_SNP_counts: the cumulative number of SNPs summed from chromosome 1 to chromosome 22 in ascending order. 
    major_minor_assignments_path: the path to the file specifying whether the major 
                                  or minor allele in each causal SNP adds 1 to the genotype. 
    betas_path: the path to the file containing one beta coefficient per row. 
    sample_size: the number of samples that have been simulated. 
    bim_SNP_names: a list of SNPs from the output bim file (same as the input bim file). 
    phenotype: an input argument specifying whether to simulate a continuous or binary phenotype. 
    output_name: name of the output bed file, which annotates the chromosome that it belongs to. 
    mean_phenotype: an input argument (float). It can be any number for continuous phenotypes, 
                    and it must be in between 0 and 1 for binary phenotypes. 
    SNP_phenotype_map_path: the path to the file specifying whether each causal SNP's phenotype map 
                            is additive, dominant, recessive, heterozygous_only, or homozygous_only.
    noise: a percentage of the mean phenotype that is the standard deviation of the 
           random gaussian noise that contributes to the simulated phenotype's values.

    Returns
    -------
    It returns a numpy array of one simulated phenotype per simulated whole genome. It only writes 
    the simulated data into plink files. It also writes the values of the inferred beta coefficients 
    and overall R^2 between the causal genotypes and the phenotype into a text file.

    """

    # imports required model components (SNPs and beta values).
    github_link = "https://github.com/EpistasisLab/regens"
    causal_SNP_IDs = open(causal_SNP_IDs_path, "r").readlines()
    try:
        betas = np.array(open(betas_path, "r").readlines()).astype(np.float64)
    except:
        print("\nerror: The beta coefficients file at " + betas_path + " is incorrectly formatted. Visit " + github_link + " for examples of correct formatting.\n")
        exit()
    if len(betas) != len(causal_SNP_IDs):
        print("\nerror: The causal_SNP_IDs and betas files must have the same number of rows. Visit " + github_link + " for examples of correct formatting.\n")
        exit()

    # imports optional model components (major/minor assignments and SNP_phenotype_maps).
    if major_minor_assignments_path != "standard":
        major_minor_assignments = open(major_minor_assignments_path, "r").readlines()
        if len(major_minor_assignments) != len(causal_SNP_IDs):
            print("\nerror: The causal_SNP_IDs and major_minor_assignments files must have the same number of rows. Visit " + github_link + " for examples of correct formatting.\n")
            exit()
    if SNP_phenotype_map_path != "standard":
        SNP_phenotype_map = open(SNP_phenotype_map_path, "r").readlines()
        if len(SNP_phenotype_map) != len(causal_SNP_IDs):
            print("\nerror: The causal_SNP_IDs and SNP_phenotype_map files must have the same number of rows. Visit " + github_link + " for examples of correct formatting.\n")
            exit()

    # simulates phenotypes based on model specifications
    feature_size = len(betas)
    features = np.zeros((sample_size, feature_size))
    for p in range(feature_size):
        feature_SNP_IDs = causal_SNP_IDs[p].strip().split('\t')
        try:
            feature_SNPs = get_feature_SNPs(feature_SNP_IDs, cumulative_SNP_counts, output_file_names, sample_size, bim_SNP_names)
        except:
            print("\nerror: The causal SNP IDs on row " + str(p + 1) + " are either incorrectly formatted or they do not exist in the input bim file:\n")
            print("Visit " + github_link + " for examples of correct formatting.\n")
        if major_minor_assignments_path != "standard":
            if np.all(np.isin(major_minor_assignments[p].strip().split('\t'), ['0', '1'])):            
                feature_major_minor_assignments = np.array(major_minor_assignments[p].strip().split('\t')).astype(np.int64)
                feature_major_minor_assignments_alt = COPY(feature_major_minor_assignments)
                feature_major_minor_assignments_alt[feature_major_minor_assignments_alt == 0] = -1
                feature_SNPs_with_assignments = (feature_SNPs - 2*feature_major_minor_assignments)*(-1*feature_major_minor_assignments_alt)
            else:
                print("\nerror: The major minor assignments on row " + str(p + 1) + " are either incorrectly formatted or they do not exist in the input bim file:\n")
                print("Visit " + github_link + " for examples of correct formatting.\n")
        if major_minor_assignments_path == "standard":
            feature_SNPs_with_assignments = feature_SNPs

        if SNP_phenotype_map_path != "standard":
            feature_SNP_phenotype_map = SNP_phenotype_map[p].strip().split('\t')
            if feature_SNPs_with_assignments.shape[1] > 1:
                for m in range(len(feature_SNP_phenotype_map)):
                    if feature_SNP_phenotype_map[m] == "recessive":
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 1] = 0
                    elif feature_SNP_phenotype_map[m] == "dominant":
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 1] = 2
                    elif feature_SNP_phenotype_map[m] == "heterozygous_only":
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 2] = 0
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 1] = 2
                    elif feature_SNP_phenotype_map[m] == "homozygous_only":
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 0] = 2
                        feature_SNPs_with_assignments[:, m][feature_SNPs_with_assignments[:, m] == 1] = 0
                    elif feature_SNP_phenotype_map[m] == "regular":
                        pass
                    else:
                        print("\nerror: all SNP_phenotype labels must be 'regular', 'recessive', 'dominant', 'heterozygous_only', or 'homozygous_only'.\n")
                        print("Visit " + github_link + " for examples of correct formatting.\n")
                        exit()
            if feature_SNPs_with_assignments.shape[1] == 1:
                for m in range(len(feature_SNP_phenotype_map)):
                    if feature_SNP_phenotype_map[m] == "recessive":
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 1] = 0
                    elif feature_SNP_phenotype_map[m] == "dominant":
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 1] = 2
                    elif feature_SNP_phenotype_map[m] == "heterozygous_only":
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 2] = 0
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 1] = 2
                    elif feature_SNP_phenotype_map[m] == "homozygous_only":
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 0] = 2
                        feature_SNPs_with_assignments[feature_SNPs_with_assignments == 1] = 0
                    elif feature_SNP_phenotype_map[m] == "regular":
                        pass
                    else:
                        print("\nerror: all SNP_phenotype labels must be 'regular', 'recessive', 'dominant', 'heterozygous_only', or 'homozygous_only'.\n")
                        print("Visit " + github_link + " for examples of correct formatting.\n")
                        exit()

        features[:, p] = np.product(feature_SNPs_with_assignments, axis = 1)
    weighted_feature_sums = np.sum(betas*features, axis = 1, keepdims = True) 
    weighted_feature_sums += np.random.normal(loc = 0, scale = noise*np.mean(weighted_feature_sums), size = weighted_feature_sums.shape)
    
    if phenotype == "binary":
        def logistic_with_unknown_intercept(intercept, weighted_feature_sums, mean_phenotype):
            disease_probabilities = 1/(1 + np.exp(-1*(weighted_feature_sums + intercept)))
            return(np.mean(disease_probabilities) - mean_phenotype)
        intercept = root(fun = logistic_with_unknown_intercept, x0 = np.array([0]), args = (weighted_feature_sums,  mean_phenotype)).x[0]
        disease_probabilities = 1/(1 + np.exp(-1*(weighted_feature_sums + intercept)))
        simulated_phenotypes = (np.random.rand(len(disease_probabilities)) <= disease_probabilities.reshape(-1)).astype(np.int8)
        model = LogisticRegression(C = 1E100, tol = 1E-100, max_iter = 1000000, solver='lbfgs').fit(features, simulated_phenotypes)

    elif phenotype == "continuous":
        def linear_with_unknown_intercept(intercept, weighted_feature_sums, mean_phenotype):
            return(np.mean(weighted_feature_sums + intercept) - mean_phenotype)
        intercept = root(fun = linear_with_unknown_intercept, x0 = np.array([0]), args = (weighted_feature_sums,  mean_phenotype)).x[0]
        simulated_phenotypes = weighted_feature_sums + intercept
        model = LinearRegression().fit(features, simulated_phenotypes)

    else:
        print("error: phenotype must be either 'binary' or 'continuous'.")
        exit()
 
    model_profile = open(output_file_names[0][:-8] + "model_profile.txt", "w")
    model_profile.write("measured R^2 of model fit: " + str(model.score(features, simulated_phenotypes)) + "\n")
    for i, b in enumerate(model.coef_[0]):
        model_profile.write("measured beta value of feature" + str(i+1) + ": " + str(b) + "\n")
    model_profile.write("measured beta value of intercept: " + str(model.intercept_[0]))
    model_profile.close()
    return(simulated_phenotypes)
  



    


from setuptools import setup
from setuptools import find_packages

regens_summary = "REGENS (REcombinatory Genome ENumeration of Subpopulations) is an open "
regens_summary += "source Python package that simulates whole genomes from real genomic "
regens_summary += "segments. REGENS recombines these segments in a way that simulates "
regens_summary += "completely new individuals while simultaneously preserving the input "
regens_summary += "genomes' linkage disequilibrium (LD) pattern with extremely high fedility. "
regens_summary += "REGENS can also simulate mono-allelic and epistatic single nucleotide variant "
regens_summary += "(SNV) effects on a continuous or binary phenotype without perturbing the simulated LD pattern."
regens_summary += "\n\nplease visit our [github page](https://github.com/EpistasisLab/regens) for more information"

setup(
    long_description = regens_summary,
    long_description_content_type = 'text/markdown',
    packages = find_packages(),
    version='0.2.0b',
    python_requires = "==3.7",
    name = 'regens',
    entry_points = {'console_scripts': ['regens = regens:main']},
    py_modules = ['regens', 'regens_library', 'regens_testers'],
    setup_requires=["numpy"],
    install_requires = ["bed-reader==0.1.1",
                        "numpy==1.16.2",
                        "pandas==1.1.2",
                        "matplotlib==3.0.3",
                        "scipy==1.2.1",
                        "scikit-learn==0.20.3"]
)
